==> ruff-0.1.12.tar.gz: ruff-0.1.12/crates/ruff_linter/src/rules/pylint/rules/unnecessary_dunder_call.rs <==
use ruff_diagnostics::{Diagnostic, Edit, Fix, FixAvailability, Violation};
use ruff_macros::{derive_message_formats, violation};
use ruff_python_ast::{self as ast, Expr};
use ruff_python_semantic::SemanticModel;
use ruff_text_size::Ranged;

use crate::checkers::ast::Checker;
use crate::rules::pylint::helpers::is_known_dunder_method;
use crate::settings::types::PythonVersion;

/// ## What it does
/// Checks for explicit use of dunder methods, like `__str__` and `__add__`.
///
/// ## Why is this bad?
/// Dunder names are not meant to be called explicitly and, in most cases, can
/// be replaced with builtins or operators.
///
/// ## Example
/// ```python
/// three = (3.0).__str__()
/// twelve = "1".__add__("2")
///
///
/// def is_greater_than_two(x: int) -> bool:
///     return x.__gt__(2)
/// ```
///
/// Use instead:
/// ```python
/// three = str(3.0)
/// twelve = "1" + "2"
///
///
/// def is_greater_than_two(x: int) -> bool:
///     return x > 2
/// ```
///
#[violation]
pub struct UnnecessaryDunderCall {
    method: String,
    replacement: Option<String>,
}

impl Violation for UnnecessaryDunderCall {
    const FIX_AVAILABILITY: FixAvailability = FixAvailability::Sometimes;

    #[derive_message_formats]
    fn message(&self) -> String {
        let UnnecessaryDunderCall {
            method,
            replacement,
        } = self;

        if let Some(replacement) = replacement {
            format!("Unnecessary dunder call to `{method}`. {replacement}.")
        } else {
            format!("Unnecessary dunder call to `{method}`")
        }
    }

    fn fix_title(&self) -> Option<String> {
        let UnnecessaryDunderCall { replacement, .. } = self;
        replacement.clone()
    }
}

/// PLC2801
pub(crate) fn unnecessary_dunder_call(checker: &mut Checker, call: &ast::ExprCall) {
    let Expr::Attribute(ast::ExprAttribute { value, attr, .. }) = call.func.as_ref() else {
        return;
    };

    // If this isn't a known dunder method, abort.
    if !is_known_dunder_method(attr) {
        return;
    }

    // If this is an allowed dunder method, abort.
    if allowed_dunder_constants(attr, checker.settings.target_version) {
        return;
    }

    // Ignore certain dunder methods used in lambda expressions.
    if allow_nested_expression(attr, checker.semantic()) {
        return;
    }

    // Ignore dunder methods used on `super`.
    if let Expr::Call(ast::ExprCall { func, .. }) = value.as_ref() {
        if checker.semantic().is_builtin("super") {
            if let Expr::Name(ast::ExprName { id, .. }) = func.as_ref() {
                if id == "super" {
                    return;
                }
            }
        }
    }

    // If the call has keywords, abort.
    if !call.arguments.keywords.is_empty() {
        return;
    }

    let mut fixed: Option<String> = None;
    let mut title: Option<String> = None;

    if let Some(dunder) = DunderReplacement::from_method(attr) {
        match (call.arguments.args.as_slice(), dunder) {
            ([], DunderReplacement::Builtin(replacement, message)) => {
                if !checker.semantic().is_builtin(replacement) {
                    return;
                }
                fixed = Some(format!(
                    "{}({})",
                    replacement,
                    checker.locator().slice(value.as_ref()),
                ));
                title = Some(message.to_string());
            }
            ([arg], DunderReplacement::Operator(replacement, message)) => {
                fixed = Some(format!(
                    "{} {} {}",
                    checker.locator().slice(value.as_ref()),
                    replacement,
                    checker.locator().slice(arg),
                ));
                title = Some(message.to_string());
            }
            ([arg], DunderReplacement::ROperator(replacement, message)) => {
                fixed = Some(format!(
                    "{} {} {}",
                    checker.locator().slice(arg),
                    replacement,
                    checker.locator().slice(value.as_ref()),
                ));
                title = Some(message.to_string());
            }
            (_, DunderReplacement::MessageOnly(message)) => {
                title = Some(message.to_string());
            }
            _ => {}
        }
    }

    let mut diagnostic = Diagnostic::new(
        UnnecessaryDunderCall {
            method: attr.to_string(),
            replacement: title,
        },
        call.range(),
    );

    if let Some(fixed) = fixed {
        diagnostic.set_fix(Fix::safe_edit(Edit::range_replacement(fixed, call.range())));
    };

    checker.diagnostics.push(diagnostic);
}

/// Return `true` if this is a dunder method that is allowed to be called explicitly.
fn allowed_dunder_constants(dunder_method: &str, target_version: PythonVersion) -> bool {
    if matches!(
        dunder_method,
        "__aexit__"
            | "__await__"
            | "__class__"
            | "__class_getitem__"
            | "__dict__"
            | "__doc__"
            | "__exit__"
            | "__getnewargs__"
            | "__getnewargs_ex__"
            | "__getstate__"
            | "__index__"
            | "__init_subclass__"
            | "__missing__"
            | "__module__"
            | "__new__"
            | "__post_init__"
            | "__reduce__"
            | "__reduce_ex__"
            | "__set_name__"
            | "__setstate__"
            | "__sizeof__"
            | "__subclasses__"
            | "__subclasshook__"
            | "__weakref__"
    ) {
        return true;
    }

    if target_version < PythonVersion::Py310 && matches!(dunder_method, "__aiter__" | "__anext__") {
        return true;
    }

    false
}

#[derive(Debug, Copy, Clone)]
enum DunderReplacement {
    /// A dunder method that is an operator.
    Operator(&'static str, &'static str),
    /// A dunder method that is a right-side operator.
    ROperator(&'static str, &'static str),
    /// A dunder method that is a builtin.
    Builtin(&'static str, &'static str),
    /// A dunder method that is a message only.
    MessageOnly(&'static str),
}

impl DunderReplacement {
    fn from_method(dunder_method: &str) -> Option<Self> {
        match dunder_method {
            "__add__" => Some(Self::Operator("+", "Use `+` operator")),
            "__and__" => Some(Self::Operator("&", "Use `&` operator")),
            "__contains__" => Some(Self::Operator("in", "Use `in` operator")),
            "__eq__" => Some(Self::Operator("==", "Use `==` operator")),
            "__floordiv__" => Some(Self::Operator("//", "Use `//` operator")),
            "__ge__" => Some(Self::Operator(">=", "Use `>=` operator")),
            "__gt__" => Some(Self::Operator(">", "Use `>` operator")),
            "__iadd__" => Some(Self::Operator("+=", "Use `+=` operator")),
            "__iand__" => Some(Self::Operator("&=", "Use `&=` operator")),
            "__ifloordiv__" => Some(Self::Operator("//=", "Use `//=` operator")),
            "__ilshift__" => Some(Self::Operator("<<=", "Use `<<=` operator")),
            "__imod__" => Some(Self::Operator("%=", "Use `%=` operator")),
            "__imul__" => Some(Self::Operator("*=", "Use `*=` operator")),
            "__ior__" => Some(Self::Operator("|=", "Use `|=` operator")),
            "__ipow__" => Some(Self::Operator("**=", "Use `**=` operator")),
            "__irshift__" => Some(Self::Operator(">>=", "Use `>>=` operator")),
            "__isub__" => Some(Self::Operator("-=", "Use `-=` operator")),
            "__itruediv__" => Some(Self::Operator("/=", "Use `/=` operator")),
            "__ixor__" => Some(Self::Operator("^=", "Use `^=` operator")),
            "__le__" => Some(Self::Operator("<=", "Use `<=` operator")),
            "__lshift__" => Some(Self::Operator("<<", "Use `<<` operator")),
            "__lt__" => Some(Self::Operator("<", "Use `<` operator")),
            "__mod__" => Some(Self::Operator("%", "Use `%` operator")),
            "__mul__" => Some(Self::Operator("*", "Use `*` operator")),
            "__ne__" => Some(Self::Operator("!=", "Use `!=` operator")),
            "__or__" => Some(Self::Operator("|", "Use `|` operator")),
            "__rshift__" => Some(Self::Operator(">>", "Use `>>` operator")),
            "__sub__" => Some(Self::Operator("-", "Use `-` operator")),
            "__truediv__" => Some(Self::Operator("/", "Use `/` operator")),
            "__xor__" => Some(Self::Operator("^", "Use `^` operator")),

            "__radd__" => Some(Self::ROperator("+", "Use `+` operator")),
            "__rand__" => Some(Self::ROperator("&", "Use `&` operator")),
            "__rfloordiv__" => Some(Self::ROperator("//", "Use `//` operator")),
            "__rlshift__" => Some(Self::ROperator("<<", "Use `<<` operator")),
            "__rmod__" => Some(Self::ROperator("%", "Use `%` operator")),
            "__rmul__" => Some(Self::ROperator("*", "Use `*` operator")),
            "__ror__" => Some(Self::ROperator("|", "Use `|` operator")),
            "__rrshift__" => Some(Self::ROperator(">>", "Use `>>` operator")),
            "__rsub__" => Some(Self::ROperator("-", "Use `-` operator")),
            "__rtruediv__" => Some(Self::ROperator("/", "Use `/` operator")),
            "__rxor__" => Some(Self::ROperator("^", "Use `^` operator")),

            "__aiter__" => Some(Self::Builtin("aiter", "Use `aiter()` builtin")),
            "__anext__" => Some(Self::Builtin("anext", "Use `anext()` builtin")),
            "__abs__" => Some(Self::Builtin("abs", "Use `abs()` builtin")),
            "__bool__" => Some(Self::Builtin("bool", "Use `bool()` builtin")),
            "__bytes__" => Some(Self::Builtin("bytes", "Use `bytes()` builtin")),
            "__complex__" => Some(Self::Builtin("complex", "Use `complex()` builtin")),
            "__dir__" => Some(Self::Builtin("dir", "Use `dir()` builtin")),
            "__float__" => Some(Self::Builtin("float", "Use `float()` builtin")),
            "__hash__" => Some(Self::Builtin("hash", "Use `hash()` builtin")),
            "__int__" => Some(Self::Builtin("int", "Use `int()` builtin")),
            "__iter__" => Some(Self::Builtin("iter", "Use `iter()` builtin")),
            "__len__" => Some(Self::Builtin("len", "Use `len()` builtin")),
            "__next__" => Some(Self::Builtin("next", "Use `next()` builtin")),
            "__repr__" => Some(Self::Builtin("repr", "Use `repr()` builtin")),
            "__reversed__" => Some(Self::Builtin("reversed", "Use `reversed()` builtin")),
            "__round__" => Some(Self::Builtin("round", "Use `round()` builtin")),
            "__str__" => Some(Self::Builtin("str", "Use `str()` builtin")),
            "__subclasscheck__" => Some(Self::Builtin("issubclass", "Use `issubclass()` builtin")),

            "__aenter__" => Some(Self::MessageOnly("Use `aenter()` builtin")),
            "__ceil__" => Some(Self::MessageOnly("Use `math.ceil()` function")),
            "__copy__" => Some(Self::MessageOnly("Use `copy.copy()` function")),
            "__deepcopy__" => Some(Self::MessageOnly("Use `copy.deepcopy()` function")),
            "__del__" => Some(Self::MessageOnly("Use `del` statement")),
            "__delattr__" => Some(Self::MessageOnly("Use `del` statement")),
            "__delete__" => Some(Self::MessageOnly("Use `del` statement")),
            "__delitem__" => Some(Self::MessageOnly("Use `del` statement")),
            "__divmod__" => Some(Self::MessageOnly("Use `divmod()` builtin")),
            "__format__" => Some(Self::MessageOnly(
                "Use `format` builtin, format string method, or f-string.",
            )),
            "__fspath__" => Some(Self::MessageOnly("Use `os.fspath` function")),
            "__get__" => Some(Self::MessageOnly("Use `get` method")),
            "__getattr__" => Some(Self::MessageOnly(
                "Access attribute directly or use getattr built-in function.",
            )),
            "__getattribute__" => Some(Self::MessageOnly(
                "Access attribute directly or use getattr built-in function.",
            )),
            "__getitem__" => Some(Self::MessageOnly("Access item via subscript")),
            "__init__" => Some(Self::MessageOnly("Instantiate class directly")),
            "__instancecheck__" => Some(Self::MessageOnly("Use `isinstance()` builtin")),
            "__invert__" => Some(Self::MessageOnly("Use `~` operator")),
            "__neg__" => Some(Self::MessageOnly("Multiply by -1 instead")),
            "__pos__" => Some(Self::MessageOnly("Multiply by +1 instead")),
            "__pow__" => Some(Self::MessageOnly("Use ** operator or `pow()` builtin")),
            "__rdivmod__" => Some(Self::MessageOnly("Use `divmod()` builtin")),
            "__rpow__" => Some(Self::MessageOnly("Use ** operator or `pow()` builtin")),
            "__set__" => Some(Self::MessageOnly("Use subscript assignment")),
            "__setattr__" => Some(Self::MessageOnly(
                "Mutate attribute directly or use setattr built-in function.",
            )),
            "__setitem__" => Some(Self::MessageOnly("Use subscript assignment")),
            "__truncate__" => Some(Self::MessageOnly("Use `math.trunc()` function")),

            _ => None,
        }
    }
}

/// Returns `true` if this is a dunder method that is excusable in a nested expression. Some
/// methods are otherwise unusable in lambda expressions and elsewhere, as they can only be
/// represented as
/// statements.
fn allow_nested_expression(dunder_name: &str, semantic: &SemanticModel) -> bool {
    semantic.current_expression_parent().is_some()
        && matches!(
            dunder_name,
            "__init__"
                | "__del__"
                | "__delattr__"
                | "__set__"
                | "__delete__"
                | "__setitem__"
                | "__delitem__"
                | "__iadd__"
                | "__isub__"
                | "__imul__"
                | "__imatmul__"
                | "__itruediv__"
                | "__ifloordiv__"
                | "__imod__"
                | "__ipow__"
                | "__ilshift__"
                | "__irshift__"
                | "__iand__"
                | "__ixor__"
                | "__ior__"
        )
}

==> ruff-0.1.13.tar.gz: ruff-0.1.13/crates/ruff_linter/src/rules/pylint/rules/unnecessary_dunder_call.rs <==
use ruff_diagnostics::{Diagnostic, Edit, Fix, FixAvailability, Violation};
use ruff_macros::{derive_message_formats, violation};
use ruff_python_ast::{self as ast, Expr};
use ruff_python_semantic::SemanticModel;
use ruff_text_size::Ranged;

use crate::checkers::ast::Checker;
use crate::rules::pylint::helpers::is_known_dunder_method;
use crate::settings::types::PythonVersion;

/// ## What it does
/// Checks for explicit use of dunder methods, like `__str__` and `__add__`.
///
/// ## Why is this bad?
/// Dunder names are not meant to be called explicitly and, in most cases, can
/// be replaced with builtins or operators.
///
/// ## Example
/// ```python
/// three = (3.0).__str__()
/// twelve = "1".__add__("2")
///
///
/// def is_greater_than_two(x: int) -> bool:
///     return x.__gt__(2)
/// ```
///
/// Use instead:
/// ```python
/// three = str(3.0)
/// twelve = "1" + "2"
///
///
/// def is_greater_than_two(x: int) -> bool:
///     return x > 2
/// ```
///
#[violation]
pub struct UnnecessaryDunderCall {
    method: String,
    replacement: Option<String>,
}

impl Violation for UnnecessaryDunderCall {
    const FIX_AVAILABILITY: FixAvailability = FixAvailability::Sometimes;

    #[derive_message_formats]
    fn message(&self) -> String {
        let UnnecessaryDunderCall {
            method,
            replacement,
        } = self;

        if let Some(replacement) = replacement {
            format!("Unnecessary dunder call to `{method}`. {replacement}.")
        } else {
            format!("Unnecessary dunder call to `{method}`")
        }
    }

    fn fix_title(&self) -> Option<String> {
        let UnnecessaryDunderCall { replacement, .. } = self;
        replacement.clone()
    }
}

/// PLC2801
pub(crate) fn unnecessary_dunder_call(checker: &mut Checker, call: &ast::ExprCall) {
    let Expr::Attribute(ast::ExprAttribute { value, attr, .. }) = call.func.as_ref() else {
        return;
    };

    // If this isn't a known dunder method, abort.
    if !is_known_dunder_method(attr) {
        return;
    }

    // If this is an allowed dunder method, abort.
    if allowed_dunder_constants(attr, checker.settings.target_version) {
        return;
    }

    // Ignore certain dunder methods used in lambda expressions.
    if allow_nested_expression(attr, checker.semantic()) {
        return;
    }

    // Ignore dunder methods used on `super`.
    if let Expr::Call(ast::ExprCall { func, .. }) = value.as_ref() {
        if checker.semantic().is_builtin("super") {
            if let Expr::Name(ast::ExprName { id, .. }) = func.as_ref() {
                if id == "super" {
                    return;
                }
            }
        }
    }

    // If the call has keywords, abort.
    if !call.arguments.keywords.is_empty() {
        return;
    }

    let mut fixed: Option<String> = None;
    let mut title: Option<String> = None;

    if let Some(dunder) = DunderReplacement::from_method(attr) {
        match (call.arguments.args.as_slice(), dunder) {
            ([], DunderReplacement::Builtin(replacement, message)) => {
                if !checker.semantic().is_builtin(replacement) {
                    return;
                }
                fixed = Some(format!(
                    "{}({})",
                    replacement,
                    checker.locator().slice(value.as_ref()),
                ));
                title = Some(message.to_string());
            }
            ([arg], DunderReplacement::Operator(replacement, message)) => {
                fixed = Some(format!(
                    "{} {} {}",
                    checker.locator().slice(value.as_ref()),
                    replacement,
                    checker.locator().slice(arg),
                ));
                title = Some(message.to_string());
            }
            ([arg], DunderReplacement::ROperator(replacement, message)) => {
                fixed = Some(format!(
                    "{} {} {}",
                    checker.locator().slice(arg),
                    replacement,
                    checker.locator().slice(value.as_ref()),
                ));
                title = Some(message.to_string());
            }
            (_, DunderReplacement::MessageOnly(message)) => {
                title = Some(message.to_string());
            }
            _ => {}
        }
    }

    let mut diagnostic = Diagnostic::new(
        UnnecessaryDunderCall {
            method: attr.to_string(),
            replacement: title,
        },
        call.range(),
    );

    if let Some(fixed) = fixed {
        diagnostic.set_fix(Fix::safe_edit(Edit::range_replacement(fixed, call.range())));
    };

    checker.diagnostics.push(diagnostic);
}

/// Return `true` if this is a dunder method that is allowed to be called explicitly.
fn allowed_dunder_constants(dunder_method: &str, target_version: PythonVersion) -> bool {
    if matches!(
        dunder_method,
        "__aexit__"
            | "__await__"
            | "__class__"
            | "__class_getitem__"
            | "__dict__"
            | "__doc__"
            | "__exit__"
            | "__getnewargs__"
            | "__getnewargs_ex__"
            | "__getstate__"
            | "__index__"
            | "__init_subclass__"
            | "__missing__"
            | "__module__"
            | "__new__"
            | "__post_init__"
            | "__reduce__"
            | "__reduce_ex__"
            | "__set_name__"
            | "__setstate__"
            | "__sizeof__"
            | "__subclasses__"
            | "__subclasshook__"
            | "__weakref__"
    ) {
        return true;
    }

    if target_version < PythonVersion::Py310 && matches!(dunder_method, "__aiter__" | "__anext__") {
        return true;
    }

    false
}

#[derive(Debug, Copy, Clone)]
enum DunderReplacement {
    /// A dunder method that is an operator.
    Operator(&'static str, &'static str),
    /// A dunder method that is a right-side operator.
    ROperator(&'static str, &'static str),
    /// A dunder method that is a builtin.
    Builtin(&'static str, &'static str),
    /// A dunder method that is a message only.
    MessageOnly(&'static str),
}

impl DunderReplacement {
    fn from_method(dunder_method: &str) -> Option<Self> {
        match dunder_method {
            "__add__" => Some(Self::Operator("+", "Use `+` operator")),
            "__and__" => Some(Self::Operator("&", "Use `&` operator")),
            "__contains__" => Some(Self::Operator("in", "Use `in` operator")),
            "__eq__" => Some(Self::Operator("==", "Use `==` operator")),
            "__floordiv__" => Some(Self::Operator("//", "Use `//` operator")),
            "__ge__" => Some(Self::Operator(">=", "Use `>=` operator")),
            "__gt__" => Some(Self::Operator(">", "Use `>` operator")),
            "__iadd__" => Some(Self::Operator("+=", "Use `+=` operator")),
            "__iand__" => Some(Self::Operator("&=", "Use `&=` operator")),
            "__ifloordiv__" => Some(Self::Operator("//=", "Use `//=` operator")),
            "__ilshift__" => Some(Self::Operator("<<=", "Use `<<=` operator")),
            "__imod__" => Some(Self::Operator("%=", "Use `%=` operator")),
            "__imul__" => Some(Self::Operator("*=", "Use `*=` operator")),
            "__ior__" => Some(Self::Operator("|=", "Use `|=` operator")),
            "__ipow__" => Some(Self::Operator("**=", "Use `**=` operator")),
            "__irshift__" => Some(Self::Operator(">>=", "Use `>>=` operator")),
            "__isub__" => Some(Self::Operator("-=", "Use `-=` operator")),
            "__itruediv__" => Some(Self::Operator("/=", "Use `/=` operator")),
            "__ixor__" => Some(Self::Operator("^=", "Use `^=` operator")),
            "__le__" => Some(Self::Operator("<=", "Use `<=` operator")),
            "__lshift__" => Some(Self::Operator("<<", "Use `<<` operator")),
            "__lt__" => Some(Self::Operator("<", "Use `<` operator")),
            "__mod__" => Some(Self::Operator("%", "Use `%` operator")),
            "__mul__" => Some(Self::Operator("*", "Use `*` operator")),
            "__ne__" => Some(Self::Operator("!=", "Use `!=` operator")),
            "__or__" => Some(Self::Operator("|", "Use `|` operator")),
            "__rshift__" => Some(Self::Operator(">>", "Use `>>` operator")),
            "__sub__" => Some(Self::Operator("-", "Use `-` operator")),
            "__truediv__" => Some(Self::Operator("/", "Use `/` operator")),
            "__xor__" => Some(Self::Operator("^", "Use `^` operator")),

            "__radd__" => Some(Self::ROperator("+", "Use `+` operator")),
            "__rand__" => Some(Self::ROperator("&", "Use `&` operator")),
            "__rfloordiv__" => Some(Self::ROperator("//", "Use `//` operator")),
            "__rlshift__" => Some(Self::ROperator("<<", "Use `<<` operator")),
            "__rmod__" => Some(Self::ROperator("%", "Use `%` operator")),
            "__rmul__" => Some(Self::ROperator("*", "Use `*` operator")),
            "__ror__" => Some(Self::ROperator("|", "Use `|` operator")),
            "__rrshift__" => Some(Self::ROperator(">>", "Use `>>` operator")),
            "__rsub__" => Some(Self::ROperator("-", "Use `-` operator")),
            "__rtruediv__" => Some(Self::ROperator("/", "Use `/` operator")),
            "__rxor__" => Some(Self::ROperator("^", "Use `^` operator")),

            "__aiter__" => Some(Self::Builtin("aiter", "Use `aiter()` builtin")),
            "__anext__" => Some(Self::Builtin("anext", "Use `anext()` builtin")),
            "__abs__" => Some(Self::Builtin("abs", "Use `abs()` builtin")),
            "__bool__" => Some(Self::Builtin("bool", "Use `bool()` builtin")),
            "__bytes__" => Some(Self::Builtin("bytes", "Use `bytes()` builtin")),
            "__complex__" => Some(Self::Builtin("complex", "Use `complex()` builtin")),
            "__dir__" => Some(Self::Builtin("dir", "Use `dir()` builtin")),
            "__float__" => Some(Self::Builtin("float", "Use `float()` builtin")),
            "__hash__" => Some(Self::Builtin("hash", "Use `hash()` builtin")),
            "__int__" => Some(Self::Builtin("int", "Use `int()` builtin")),
            "__iter__" => Some(Self::Builtin("iter", "Use `iter()` builtin")),
            "__len__" => Some(Self::Builtin("len", "Use `len()` builtin")),
            "__next__" => Some(Self::Builtin("next", "Use `next()` builtin")),
            "__repr__" => Some(Self::Builtin("repr", "Use `repr()` builtin")),
            "__reversed__" => Some(Self::Builtin("reversed", "Use `reversed()` builtin")),
            "__round__" => Some(Self::Builtin("round", "Use `round()` builtin")),
            "__str__" => Some(Self::Builtin("str", "Use `str()` builtin")),
            "__subclasscheck__" => Some(Self::Builtin("issubclass", "Use `issubclass()` builtin")),

            "__aenter__" => Some(Self::MessageOnly("Invoke context manager directly")),
            "__ceil__" => Some(Self::MessageOnly("Use `math.ceil()` function")),
            "__copy__" => Some(Self::MessageOnly("Use `copy.copy()` function")),
            "__deepcopy__" => Some(Self::MessageOnly("Use `copy.deepcopy()` function")),
            "__del__" => Some(Self::MessageOnly("Use `del` statement")),
            "__delattr__" => Some(Self::MessageOnly("Use `del` statement")),
            "__delete__" => Some(Self::MessageOnly("Use `del` statement")),
            "__delitem__" => Some(Self::MessageOnly("Use `del` statement")),
            "__divmod__" => Some(Self::MessageOnly("Use `divmod()` builtin")),
            "__format__" => Some(Self::MessageOnly(
                "Use `format` builtin, format string method, or f-string",
            )),
            "__fspath__" => Some(Self::MessageOnly("Use `os.fspath` function")),
            "__get__" => Some(Self::MessageOnly("Use `get` method")),
            "__getattr__" => Some(Self::MessageOnly(
                "Access attribute directly or use getattr built-in function",
            )),
            "__getattribute__" => Some(Self::MessageOnly(
                "Access attribute directly or use getattr built-in function",
            )),
            "__getitem__" => Some(Self::MessageOnly("Access item via subscript")),
            "__init__" => Some(Self::MessageOnly("Instantiate class directly")),
            "__instancecheck__" => Some(Self::MessageOnly("Use `isinstance()` builtin")),
            "__invert__" => Some(Self::MessageOnly("Use `~` operator")),
            "__neg__" => Some(Self::MessageOnly("Multiply by -1 instead")),
            "__pos__" => Some(Self::MessageOnly("Multiply by +1 instead")),
            "__pow__" => Some(Self::MessageOnly("Use ** operator or `pow()` builtin")),
            "__rdivmod__" => Some(Self::MessageOnly("Use `divmod()` builtin")),
            "__rpow__" => Some(Self::MessageOnly("Use ** operator or `pow()` builtin")),
            "__set__" => Some(Self::MessageOnly("Use subscript assignment")),
            "__setattr__" => Some(Self::MessageOnly(
                "Mutate attribute directly or use setattr built-in function",
            )),
            "__setitem__" => Some(Self::MessageOnly("Use subscript assignment")),
            "__truncate__" => Some(Self::MessageOnly("Use `math.trunc()` function")),

            _ => None,
        }
    }
}

/// Returns `true` if this is a dunder method that is excusable in a nested expression. Some
/// methods are otherwise unusable in lambda expressions and elsewhere, as they can only be
/// represented as
/// statements.
fn allow_nested_expression(dunder_name: &str, semantic: &SemanticModel) -> bool {
    semantic.current_expression_parent().is_some()
        && matches!(
            dunder_name,
            "__init__"
                | "__del__"
                | "__delattr__"
                | "__set__"
                | "__delete__"
                | "__setitem__"
                | "__delitem__"
                | "__iadd__"
                | "__isub__"
                | "__imul__"
                | "__imatmul__"
                | "__itruediv__"
                | "__ifloordiv__"
                | "__imod__"
                | "__ipow__"
                | "__ilshift__"
                | "__irshift__"
                | "__iand__"
                | "__ixor__"
                | "__ior__"
        )
}
